C Program to read in the weights and the hessian matrix from output files
C produced by MacKay's bigback program and use them to calculate the outputs
C and errors associated with a new, user-suppplied parameter set.
C
C ND   - no. parameters
C INM  - no. models in the committee
C X0   - user-supplied parameter set
C XN   - normalised parameter set
C PMAX(i) - maximum value of parameter i - used for normalisation
C PMIN(i) - minimum value of parameter i - used for normalisation
C
C The data must be normalised so that the maxima and minima values given in
C PMAX and PMIN for each variable lie at +0.5 and -0.5 respectively.
C
C IHU(j) -  no. hidden units in jth network
C SNU(j) - sigma_nu or sigma_noise (taken from the report file) for jth network
C
C ANS0 - Predicted value (unnormalised) from one neural network.
C ERR0 - Size of error bar.
C ANS0P = ANS0+ERR0
C ANS0M = ANS0-ERR0
C ANS1 - Mean predicted value for a nerual network committee.
C ANS1P = ANS1+ERR1
C ANS1M = ANS1-ERR1
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER(I-N)
      DOUBLE PRECISION SNU(50),PMAX(20),PMIN(20),XN(20),X0(20)
      INTEGER IHU(50)
      CHARACTER*9 WGTFILE(50),BLANK
C
      BLANK  = '         '
      DO 100 L1=1,50
         WGTFILE(L1) = BLANK
  100 CONTINUE
C
C Call subroutine to set the maximum and minimum parameter values for the
C model and set the number of input parameters and hidden variables etc.
C
      CALL MODEL1(ND,INM,IHU,SNU,PMAX,PMIN,WGTFILE)
C
C Read in the parameters (not normalised) and store in X0
C
      READ(*,*) (X0(J),J=1,ND)
C
C Normalise the parameters. Put normalised parameters in XN
C
      CALL NORM(ND,X0,XN,PMAX,PMIN)
      ANSR = 0D0
      ANSS = 0D0
      ERR1 = 0D0
      DO 120 L1=1,INM
         CALL GETV(ND,IHU(L1),SNU(L1),PMAX,PMIN,XN,WGTFILE(L1),
     +                                         ERR0,ANS0,ANS0P,ANS0M)
         ANSR = ANSR + ANS0
         ANSS = ANSS + ANS0*ANS0
         ERR1 = ERR1 + ERR0*ERR0
  120 CONTINUE
      ERR1 = ( ERR1 + ANSS - ANSR*ANSR/INM )/INM
      ERR1 = SQRT(ERR1)
      ANS1 = ANSR / INM
      ANS1M = ANS1 - ERR1
      ANS1P = ANS1 + ERR1
C
C Write out the results
C
      WRITE(*,1)
      WRITE(*,2) ANS1,ERR1,ANS1M,ANS1P
      STOP
    1 FORMAT(///'   VALUE   ERROR        RANGE')
    2 FORMAT(2X,F6.2,2X,F6.3,2X,F6.2,' - ',F6.2)
      END
C
C***************************************************************************
C
C Subroutine to calculate the output value and error from 1 model
C
C WIJ  - array containing weights Wij 
C WI   - array containing weights for the hidden nodes
C TI   - array containing biases 
C T0   - the last single bias
C LUH  - the LU decomposition of the hessian matrix
C IHU  - array containing no. hidden units
C SNU  - sigma_nu or sigma_noise (taken from the report file)
C
      SUBROUTINE GETV(ND,IHU,SNU,PMAX,PMIN,XN,WGTFILE,ERR1,ANS1,
     +                                                    ANS1P,ANS1M)
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER(I-N)
      DOUBLE PRECISION WIJ(100,20),WI(30),TI(30),LUH(200,200)
      DOUBLE PRECISION PMAX(20),PMIN(20),XN(20),G(200)
      INTEGER INDX(200)
      CHARACTER  WGTFILE*9, LUFILE*12
C
      IUN  = 10
      LUFILE(1:9)   = WGTFILE
      LUFILE(10:12) = '.lu'
C
C Read in the weights from the weights file.
C
      OPEN(IUN,FILE=WGTFILE)
      CALL WGTREAD(IUN,IHU,ND,TI,WIJ,T0,WI)
      CLOSE(IUN)
C
C Read in the inverse hessian matrix of dimension IHE
C
      IHE = (ND+1)*IHU + IHU + 1
      OPEN(IUN,FILE=LUFILE)
      CALL LUHREAD(IUN,IHE,LUH)
      READ(IUN,*) (INDX(J),J=1,IHE)
      CLOSE(IUN)
C
C Evaluate the g-matrix (differentials of the output wrt the weights and
C biases)
C
      CALL GCALC(ND,IHU,G,TI,WIJ,WI,XN)
C
C Evaluate the size of the error bar from the equation LUH.Ginv.s^2 = G
C Total error^2 is s^2 + s_nu^2.
C
      CALL ERRCALC(ND,IHU,G,LUH,INDX,ERR)
      ERRT = ERR + SNU*SNU
      ERR1 = SQRT(ERRT)*(PMAX(ND+1)-PMIN(ND+1))
C
C Evaluate the output value Y itself and then unnormalise it.
C
      CALL YCALC(ND,IHU,Y,TI,WIJ,T0,WI,XN)
      CALL UNNORM(ND,YUN,Y,PMAX,PMIN)
      ANS1  = YUN
      ANS1P = ANS1 + ERR1
      ANS1M = ANS1 - ERR1
      IF ( ANS1M.LT.0D0 ) ANS1M = 0D0
      RETURN
      END
C
C***************************************************************************
C
      SUBROUTINE MODEL1(ND,INM,IHU,SNU,PMAX,PMIN,WGTFILE)
      IMPLICIT DOUBLE PRECISION (A-H,O-Z),INTEGER(I-N)
      DOUBLE PRECISION PMAX(20),PMIN(20),SNU(50)
      CHARACTER*9 WGTFILE(50)
      INTEGER IHU(50)
C
C ND - number of input data items for the network
C INM - number of models in the committee
C IHU - array containing number of hidden units for each model
C SNU - array containing sigma_nu for each model
C PMIN - array containing minimum parameter value (for nomalising the inputs)
C PMAX - array containing maximum parameter value (for nomalising the inputs)
C WGTFILE - array containing name of weights file - IMPORTANT - keep to the
C           same number of characters as shown below.
C
      ND      = 4
      INM     = 3
      IHU(1)  = 1
      IHU(2)  = 2
      IHU(3)  = 5
      SNU(1)  = 0.081727D0
      SNU(2)  = 0.081665D0
      SNU(3)  = 0.081618D0
      PMIN(1) = 0.8D0
      PMIN(2) = 1.44D0
      PMIN(3) = 588D0
      PMIN(4) = 8D0
      PMIN(5) = 0.93D0
      PMAX(1) = 1D0
      PMAX(2) = 3.99D0
      PMAX(3) = 1729D0
      PMAX(4) = 50D0
      PMAX(5) = 3.7459D0
      WGTFILE(1) = 'w/wgts101'
      WGTFILE(2) = 'w/wgts102'
      WGTFILE(3) = 'w/wgts103'
      RETURN
      END
C
C***************************************************************************
C
C Subroutine to normalise the parameters in X and put them in XN
C
C ND - number of parameters
C PMAX - maximum value
C PMIN - minimum value
C
      SUBROUTINE NORM(ND,X,XN,PMAX,PMIN)
      IMPLICIT DOUBLE PRECISION (A-H,O-Z),INTEGER(I-N)
      DOUBLE PRECISION X(20),XN(20),PMAX(20),PMIN(20)
C
      DO 100 L1=1,ND
         XN(L1) = (X(L1)-PMIN(L1))/(PMAX(L1)-PMIN(L1)) - 0.5D0
  100 CONTINUE
      RETURN
      END
C
C***************************************************************************
C
C Subroutine to unnormalise the output parameter in XN and put it into X
C
C ND - number of parameters
C PMAX - maximum value
C PMIN - minimum value
C
      SUBROUTINE UNNORM(ND,X,XN,PMAX,PMIN)
      IMPLICIT DOUBLE PRECISION (A-H,O-Z),INTEGER(I-N)
      DOUBLE PRECISION PMAX(20),PMIN(20)
C
      X = (XN+0.5D0)*(PMAX(ND+1)-PMIN(ND+1))+PMIN(ND+1)
      RETURN
      END
C
C***************************************************************************
C
C Subroutine to read in the weights and biases from the weights file

      SUBROUTINE WGTREAD(IUN,IHU,ND,TI,WIJ,T0,WI)
      IMPLICIT DOUBLE PRECISION (A-H,O-Z),INTEGER(I-N)
      DOUBLE PRECISION WIJ(100,20),WI(30),TI(30),T0
C
      READ(IUN,*) (TI(I),(WIJ(I,J),J=1,ND),I=1,IHU),T0,(WI(K),K=1,IHU)   
      RETURN
      END
C
C***************************************************************************
C
C Subroutine to read in the lu decomposition of the hessian from the .lu file

      SUBROUTINE LUHREAD(IUN,IHE,LUH)
      IMPLICIT DOUBLE PRECISION (A-H,O-Z),INTEGER(I-N)
      DOUBLE PRECISION LUH(200,200)
C
      DO 100 L1=1,IHE
         READ(IUN,*) (LUH(L1,J),J=1,IHE)
100   CONTINUE
      RETURN
      END
C
C***************************************************************************
C
C Subroutine to calculate the  g-vector (differentials of the output wrt the 
C weights and biases) and the inverse of the g-vector.
C
C This assumes that the NN output function is a tanh function, as in Harry's
C paper.
C
      SUBROUTINE GCALC(ND,IHU,G,TI,WIJ,WI,XN)
      IMPLICIT DOUBLE PRECISION (A-H,O-Z),INTEGER(I-N)
      DOUBLE PRECISION WIJ(100,20),WI(30),TI(30),XN(20),G(200)
C
      IT = IHU*(ND+1)+1
      DO 120 L1=1,IHU
         I = (L1-1)*(ND+1)
         SUM = 0.0D0
         DO 100 L2=1,ND
            SUM = SUM + WIJ(L1,L2)*XN(L2)
  100    CONTINUE
         SUM   = SUM + TI(L1)
         EXPP  = DEXP(SUM)
         EXPM  = DEXP(-SUM)
         SCOSH = 0.5D0*(EXPP+EXPM)
         SSINH = 0.5D0*(EXPP-EXPM)
         G(1+I)= WI(L1)/(SCOSH*SCOSH)
         DO 110 L2=1,ND
            G(L2+1+I) = XN(L2)*G(1+I)
  110    CONTINUE
         G(IT+L1) = SSINH/SCOSH
  120 CONTINUE
      G(IT) = 1D0
      RETURN
      END
C
C***************************************************************************
C
C Subroutine to evaluate the error from the equation LUH.Ginv.s^2=G
C
      SUBROUTINE ERRCALC(ND,IHU,G,LUH,INDX,ERR)
      IMPLICIT DOUBLE PRECISION (A-H,O-Z),INTEGER(I-N)
      DOUBLE PRECISION G(200),LUH(200,200),B(200)
      INTEGER INDX(200)
C
C Find size of matrix
C
      ID = (ND+1)*IHU + IHU + 1
      DO 100 L1=1,ID
         B(L1) = G(L1)
  100 CONTINUE
C
C Call LUBKSB to find Ginv.s^2
C
      CALL LUBKSB(LUH,ID,INDX,B)
      ERR = 0D0
      DO 110 L1=1,ID
         ERR = ERR + G(L1)*B(L1)
  110 CONTINUE
      RETURN
      END
C
C***************************************************************************
C
C Subroutine to calculate the output value from the weights

      SUBROUTINE YCALC(ND,IHU,Y,TI,WIJ,T0,WI,XN)
      IMPLICIT DOUBLE PRECISION (A-H,O-Z),INTEGER(I-N)
      DOUBLE PRECISION WIJ(100,20),WI(30),TI(30),XN(20)
C
      Y = 0D0
      DO 110 L1=1,IHU
         SUM = 0.0D0
         DO 100 L2=1,ND
            SUM = SUM + WIJ(L1,L2)*XN(L2)
  100    CONTINUE
         SUM   = SUM + TI(L1)
         EXPP  = DEXP(SUM)
         EXPM  = DEXP(-SUM)
         SCOSH = 0.5D0*(EXPP+EXPM)
         SSINH = 0.5D0*(EXPP-EXPM)
         STANH = SSINH/SCOSH
         Y = Y + WI(L1)*STANH
  110 CONTINUE
      Y = Y + T0
      RETURN
      END


